#ifndef VFH_NEAREST_NEIGHBOR_H
#define VFH_NEAREST_NEIGHBOR_H

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <math.h>

#include <Eigen/Dense>
#include <Eigen/Geometry>

#include <flann/flann.h>
#include <flann/io/hdf5.h>

#include <boost/thread/thread.hpp>
#include <boost/filesystem.hpp>

#include <pcl/features/vfh.h>
#include <pcl/point_types.h>
#include <pcl/io/pcd_io.h>
#include <pcl/registration/icp.h>
#include <pcl/common/transforms.h>


using namespace std;

typedef std::vector<std::vector<double> > Mat;

class Estimate6DOF
{
  public:
    float q_[4], xyz_[3];
    Eigen::Matrix4d align_transformation, transformation;

    static std::string numberToString(float x, float y, float z,
      float q1, float q2, float q3, float q4);
    static void print4x4Matrix (const Eigen::Matrix4d &matrix, bool flag_milimeter = false);
    static Mat InputMatrix (const std::string &pose_filename);
    int estimate6DOF (pcl::PointCloud<pcl::PointXYZ>::Ptr &nearest_model_cloud,
      pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_tgt,
      Eigen::Vector4d &position_base_at_camera, Eigen::Matrix4d &camera_to_robot);
    Eigen::Matrix4d angle2pose(int angle_z, int angle_y);
};

#endif
