#ifndef PCL_REGISTRATION_H
#define PCL_REGISTRATION_H

#include <pcl/io/pcd_io.h>
#include <pcl/point_types.h>
#include <pcl/point_cloud.h>
#include <pcl/registration/ia_ransac.h>
#include <pcl/registration/icp.h>
#include <pcl/common/transforms.h>
#include <pcl/common/time.h>
#include <pcl/common/centroid.h>
#include <pcl/features/fpfh.h>
#include <pcl/features/normal_3d.h>

#include <Eigen/Eigen>




class SAC_IA
{
  public:
    double score_;
    Eigen::Matrix4d sac_trans_;
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_sac_;

    void SAC_IA_Registration (pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_src, pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_tgt, double radius_normal=0.03, double radius_fpfh=0.05);
};


class ICP
{
  public:
    double score_;
    Eigen::Matrix4d icp_trans_;
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_icp_;

    void ICP_Registration (pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_src, pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_tgt,  Eigen::Matrix4d transform_init=Eigen::Matrix4d::Identity(), double MaxCorrespondenceDistance=0.1, int MaximumIterations=300, double TransformationEpsilon=1e-10, double EuclideanFitnessEpsilon=0.01 );
};

class TRANS_ICP
{
  public:
    double score_;
    Eigen::Matrix4d trans_;
    Eigen::Matrix4d transicp_trans_;
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_transicp_;

    void trans_icp (pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_src, pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_tgt,  double MaxCorrespondenceDistance=0.1, int MaximumIterations=300, double TransformationEpsilon=1e-10, double EuclideanFitnessEpsilon=0.01 );
};

class SAC_ICP
{
  public:
    double score_;
    Eigen::Matrix4d sac_;
    Eigen::Matrix4d sacicp_trans_;
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_sacicp_;

    void sac_icp (pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_src, pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_tgt, double radius_normal=0.03, double radius_fpfh=0.05, double MaxCorrespondenceDistance=0.1, int MaximumIterations=300, double TransformationEpsilon=1e-10, double EuclideanFitnessEpsilon=0.01 );
};

#endif

