#include "../include/VFHNearestNeighbor.h"

#define DEBUG
#define _DEBUG

using namespace std;

void VFHNearestNeighbor::nearestKSearch (flann::Index<flann::ChiSquareDistance<float> > &index, const vfh_model &model,
  int k, flann::Matrix<int> &indices, flann::Matrix<float> &distances)
{
  // Query point
  flann::Matrix<float> p = flann::Matrix<float>(new float[model.second.size ()], 1, model.second.size ());
  memcpy (&p.ptr ()[0], &model.second[0], p.cols * p.rows * sizeof (float));

  indices = flann::Matrix<int>(new int[k], 1, k);
  distances = flann::Matrix<float>(new float[k], 1, k);
  index.knnSearch (p, indices, distances, k, flann::SearchParams (512));
  delete[] p.ptr ();
}

bool VFHNearestNeighbor::loadFileList (std::vector<vfh_model> &models, const std::string &filename)
{
  ifstream fs;
  fs.open (filename.c_str ());
  if (!fs.is_open () || fs.fail ())
    return (false);

  std::string line;
  while (!fs.eof ())
   {
    getline (fs, line);
    if (line.empty ())
     continue;
    vfh_model m;
    m.first = line;
    models.push_back (m);
  }
  fs.close ();
  return (true);
}

int VFHNearestNeighbor::estimate6DOF(vector< pcl::PointCloud<pcl::VFHSignature308> > &vfh_vector,
  vector< pcl::PointCloud<pcl::PointXYZ> > &cluster_scene_vector)
{
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZ>);
  std::vector<float> neighbor_distance;
  std::vector<std::string> neighbor_model_name_vec;
  vfh_model histogram;
  flann::Matrix<int> k_indices;
  flann::Matrix<float> k_distances;
  flann::Matrix<float> data;

  //get the vfh from the vfh
  for (size_t t = 0; t < vfh_vector.size(); ++t)
  {
    std::stringstream s1;
    s1 << "cluster_";
    histogram.first = s1.str ();
    histogram.second.resize (308);
    for (int ti = 0; ti < 308; ++ti)
      histogram.second[ti] = vfh_vector[t].points[0].histogram[ti];

    std::string kdtree_idx_file_name = "kdtree.idx";
    std::string training_data_h5_file_name = "training_data.h5";
    std::string training_data_list_file_name = "training_data.list";
    std::vector<vfh_model> models;
    // Check if the data has already been saved to disk
    if (!boost::filesystem::exists ("training_data.h5") || !boost::filesystem::exists ("training_data.list"))
    {
      pcl::console::print_error ("Could not find training data models files %s and %s!\n",
      training_data_h5_file_name.c_str (), training_data_list_file_name.c_str ());
      return (-1);
    }
    else
    {
      loadFileList (models, training_data_list_file_name);
      flann::load_from_file (data, training_data_h5_file_name, "training_data");
    }

    int k = 1;//only use the first one, so k=1

    // Check if the tree index has already been saved to disk
    if (!boost::filesystem::exists (kdtree_idx_file_name))
    {
      pcl::console::print_error ("Could not find kd-tree index in file %s!", kdtree_idx_file_name.c_str ());
      return (-1);
    }
    else
    {
      flann::Index<flann::ChiSquareDistance<float> > index (data, flann::SavedIndexParams ("kdtree.idx"));
      index.buildIndex ();
      nearestKSearch (index, histogram, k, k_indices, k_distances);
    }

    //std::string model_name = models.at (k_indices[0][0]).first.c_str ();//only use the nearest one "k_indices[0][0])"
    neighbor_distance.push_back (k_distances[0][0]);
    //neighbor_model_name_vec.push_back (model_name);//push the nearest model name in the name vector
  }

  std::vector<float>::iterator mindistance = min_element (neighbor_distance.begin(), neighbor_distance.end());
  int obj_index = distance (neighbor_distance.begin(), mindistance);
  cout << "The obj_index is " << obj_index << endl;

  cloud_obj.reset(new pcl::PointCloud<pcl::PointXYZ>);
  *cloud_obj = cluster_scene_vector[obj_index];
  pcl::PointCloud<pcl::VFHSignature308>::Ptr obj_vfh(new pcl::PointCloud<pcl::VFHSignature308>);
  *obj_vfh = vfh_vector[obj_index];

/*-------select best model by comparing icp result----*/
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_neigh(new pcl::PointCloud<pcl::PointXYZ>);
  TRANS_ICP transicp;
  std::vector<Eigen::Matrix4d, Eigen::aligned_allocator<Eigen::Matrix4d> > neighbor_matrix;
  std::vector<double> vector_score;

  std::stringstream s2;
  s2 << "cluster_";
  histogram.first = s2.str ();
  histogram.second.resize (308);
  for (int ti = 0; ti < 308; ++ti)
    histogram.second[ti] = obj_vfh->points[0].histogram[ti];

  std::string kdtree_idx_file_name = "kdtree.idx";
  std::string training_data_h5_file_name = "training_data.h5";
  std::string training_data_list_file_name = "training_data.list";
  std::vector<vfh_model> models;
  // Check if the data has already been saved to disk
  if (!boost::filesystem::exists ("training_data.h5") || !boost::filesystem::exists ("training_data.list"))
  {
    pcl::console::print_error ("Could not find training data models files %s and %s!\n",
    training_data_h5_file_name.c_str (), training_data_list_file_name.c_str ());
    return (-1);
  }
  else
  {
    loadFileList (models, training_data_list_file_name);
    flann::load_from_file (data, training_data_h5_file_name, "training_data");
  }

  int k = 5;//only use the first one, so k=1

  // Check if the tree index has already been saved to disk
  if (!boost::filesystem::exists (kdtree_idx_file_name))
  {
    pcl::console::print_error ("Could not find kd-tree index in file %s!", kdtree_idx_file_name.c_str ());
    return (-1);
  }
  else
  {
    flann::Index<flann::ChiSquareDistance<float> > index (data, flann::SavedIndexParams ("kdtree.idx"));
    index.buildIndex ();
    nearestKSearch (index, histogram, k, k_indices, k_distances);
  }

  for (int i=0;i<k;++i)
  {
      std::string model_name = models.at (k_indices[0][i]).first.c_str ();
      boost::replace_last (model_name, "_vfh", "");
      neighbor_model_name_vec.push_back (model_name);
      cout << "Alternative neighbor : " << model_name << endl;

      if (pcl::io::loadPCDFile<pcl::PointXYZ> (model_name, *cloud_neigh) == -1) //* load the file
      {
        PCL_ERROR ("Couldn't read file *.pcd \n");
      }

      transicp.trans_icp(cloud_neigh,cloud_obj,0.1);
      neighbor_matrix.push_back(transicp.transicp_trans_);
      vector_score.push_back(transicp.score_);
  }

  std::vector<double>::iterator minscore = min_element (vector_score.begin(), vector_score.end());
  int neighbor_index = distance (vector_score.begin(), minscore);
  //icp_matrix_ = neighbor_matrix[neighbor_index];
  neighbor_model_name = neighbor_model_name_vec[neighbor_index];//this is the name of the nearest model name
  cout << "The neighbor is " << neighbor_model_name << endl;

  return (0);
}
