#include<ros/ros.h>
#include<pcl_ros/point_cloud.h>
#include<pcl_conversions/pcl_conversions.h>
#include<sensor_msgs/PointCloud2.h>
#include <std_msgs/String.h>
#include <geometry_msgs/Pose.h>
#include <geometry_msgs/PoseStamped.h>
#include <math.h>

#include <pcl/visualization/cloud_viewer.h>
#include <pcl/visualization/pcl_visualizer.h>
#include <iostream>
#include <pcl/io/pcd_io.h>
#include<pcl/point_cloud.h>
#include <pcl/point_types.h>

#include <boost/filesystem.hpp>

#include "PCLRegistration.h"
#include "loadfile.h"
#include "Estimate6DOF.h"


#define DEBUG

TRANS_ICP transicp;
Estimate6DOF e6;

ros::Publisher pub;
std::string neighborname;
Eigen::Matrix4d neigh2obj;
Eigen::Matrix4d obj_pose;
Eigen::Matrix4d camera2world;
Eigen::Matrix4d world2camera;
Eigen::Matrix4d robot2world;
geometry_msgs::Pose pose;
geometry_msgs::PoseStamped pose_pub;

Eigen::Matrix3d  pose_reg (Eigen::Matrix3d obj_pose)
{
    Eigen::Vector3d z1,z2,axis;
    double angle;
    z1 = obj_pose.col(2);
    z2<<0,0,1;
    axis=z1.cross(z2);
    axis.normalize();
    angle=acos(z1.dot(z2));
#ifdef DEBUG
    std::cout<<"Angle:"<<angle<<std::endl;
    std::cout<<"Axis:"<<axis<<std::endl;
#endif
    Eigen::AngleAxisd reg(angle, axis);
    Eigen::Matrix3d t_R = reg.matrix();
    obj_pose = t_R*obj_pose;
    return obj_pose;
}

void
cloud_cb (const sensor_msgs::PointCloud2ConstPtr& cloud_msg)
{
    ros::param::set("~get_neighborname",false);

    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_neigh(new pcl::PointCloud<pcl::PointXYZ>);

    pcl::fromROSMsg (*cloud_msg, *cloud);

    boost::replace_last (neighborname, "_vfh", "");
#ifdef DEBUG
    std::cout<<neighborname<<std::endl;
#endif
    if (pcl::io::loadPCDFile<pcl::PointXYZ> (neighborname, *cloud_neigh) == -1) //* load the file
    {
      PCL_ERROR ("Couldn't read file *.pcd \n");
    }

    transicp.trans_icp(cloud_neigh,cloud,0.1);

    if(transicp.score_ >= 0.0001)
    {
        ROS_INFO("Registration ERROR! Please try again.");
        return;
    }

    neigh2obj=transicp.transicp_trans_;

    int angle_z,angle_y;
    Loadfile::getNum(neighborname, angle_z, angle_y);

#ifdef DEBUG
      std::cout<<"Z:"<<angle_z<<"Y:"<<angle_y<<std::endl;
#endif

    obj_pose = e6.angle2pose(angle_z,angle_y);//model pose


    camera2world << 1, -0.0051, 0.0065, 0.208786,
                    0.0029, 0.9523, 0.3051, 0.0987192,
                    -0.0078, -0.3051, 0.9523, 0.821508,
                    0, 0, 0, 1;
    world2camera = camera2world.inverse();
    obj_pose = world2camera*neigh2obj*obj_pose;//in world frame
#ifdef DEBUG
    std::cout<<"POSE in cali frame:"<<std::endl<<obj_pose<<std::endl;
#endif

    robot2world << 0,-1,0, 0.0875,
                   -1,0,0,-0.2055,
                   0,0,-1,-0.1-0.004,
                   0,0,0,1;
    obj_pose = robot2world*obj_pose;   

    std::cout<<"POSE:"<<std::endl<<obj_pose<<std::endl;


    pose.position.x = obj_pose(0,3);
    pose.position.y = obj_pose(1,3);
    pose.position.z = obj_pose(2,3);


    Eigen::Matrix3d matrix_to_q;
    for (int mi = 0; mi < 3; mi++)
      for (int ni = 0; ni < 3; ni++)
        matrix_to_q(mi, ni) = obj_pose (mi, ni);

//pose regulate to 3OF
    Eigen::Matrix3d  pose_reg (Eigen::Matrix3d);
    matrix_to_q = pose_reg ( matrix_to_q);
    std::cout<<"POSE after reg:"<<std::endl<<matrix_to_q<<std::endl;

    Eigen::Quaterniond q (matrix_to_q);

    pose.orientation.x = q.x();
    pose.orientation.y = q.y();
    pose.orientation.z = q.z();
    pose.orientation.w = q.w();

    std::cout<<"publish POSE:"<<std::endl<<pose<<std::endl;
    pose_pub.pose = pose;
    pub.publish(pose_pub);

#ifdef DEBUG
    // Initializing point cloud visualizer
    boost::shared_ptr<pcl::visualization::PCLVisualizer>
    viewer_final (new pcl::visualization::PCLVisualizer ("3D Viewer"));

    viewer_final->setBackgroundColor (0, 0, 0);

    // Coloring and visualizing target cloud (red).
    pcl::visualization::PointCloudColorHandlerCustom<pcl::PointXYZ>
    target_color (cloud, 255, 0, 0);
    viewer_final->addPointCloud<pcl::PointXYZ> (cloud, target_color, "target cloud");
    viewer_final->setPointCloudRenderingProperties (pcl::visualization::PCL_VISUALIZER_POINT_SIZE,
                                                    1, "target cloud");

    // Coloring and visualizing transformed input cloud (green).
    pcl::visualization::PointCloudColorHandlerCustom<pcl::PointXYZ>
    output_color (transicp.cloud_transicp_, 0, 255, 0);
    viewer_final->addPointCloud<pcl::PointXYZ> (transicp.cloud_transicp_, output_color, "output cloud");
    viewer_final->setPointCloudRenderingProperties (pcl::visualization::PCL_VISUALIZER_POINT_SIZE,
                                                    1, "output cloud");

    // Starting visualizer
    viewer_final->addCoordinateSystem (1.0,0);
    viewer_final->initCameraParameters ();

    while (!viewer_final->wasStopped ())
    {
      viewer_final->spinOnce (100);
      boost::this_thread::sleep (boost::posix_time::microseconds (100000));
    }

#endif
}

void
string_cb (const std_msgs::String::ConstPtr& msg)
{
    neighborname = msg->data;
#ifdef DEBUG
    std::cout<<msg->data<<"0"<<std::endl;
#endif
    ros::param::set("~get_neighborname",true);
}

int
main (int argc, char** argv)
{
  // Initialize ROS
  ros::init (argc, argv, "pose_estimation");
  ros::NodeHandle nh;

  pub = nh.advertise<geometry_msgs::PoseStamped> ("pose", 1);


  // Create a ROS subscriber for the input point cloud
  ros::Subscriber sub_neighbor_name = nh.subscribe<std_msgs::String> ("neighbor_name", 1, string_cb);
  ros::Subscriber sub_obj = nh.subscribe<sensor_msgs::PointCloud2> ("cloud_obj", 1, cloud_cb);


  ros::AsyncSpinner spinner(3);
  spinner.start();
  ros::waitForShutdown();

  return 0;
}

