#include<ros/ros.h>
#include<pcl_ros/point_cloud.h>
#include<pcl_conversions/pcl_conversions.h>
#include<sensor_msgs/PointCloud2.h>
#include <std_msgs/String.h>

#include <pcl/visualization/cloud_viewer.h>
#include <iostream>
#include <pcl/io/pcd_io.h>
#include<pcl/point_cloud.h>
#include <pcl/point_types.h>
#include <pcl/filters/filter.h>
#include <pcl/ModelCoefficients.h>
#include <pcl/filters/extract_indices.h>
#include <pcl/features/normal_3d.h>
#include <pcl/kdtree/kdtree.h>
#include <pcl/sample_consensus/method_types.h>
#include <pcl/sample_consensus/model_types.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/segmentation/extract_clusters.h>

#include <PCLSegmentation.h>
#include <PCLFeatures.h>
#include <VFHNearestNeighbor.h>

#define DEBUG

ros::Publisher pub_obj;
ros::Publisher pub_neighbor;

ExtractClusterOnPlane ex;
Feature feature;
VFHNearestNeighbor vnn;

char filename[32];

void
cloud_cb (const sensor_msgs::PointCloud2ConstPtr& cloud_msg)
{
    bool seg_flag;
    ros::param::get("~seg_flag",seg_flag);
    if(seg_flag==true)
    {
      pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
      pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_seg(new pcl::PointCloud<pcl::PointXYZ>);
      pcl::PointCloud<pcl::PointXYZ>::Ptr cluster(new pcl::PointCloud<pcl::PointXYZ>);
      std::vector< pcl::PointCloud<pcl::PointXYZ> > cluster_vec;
      std::vector< pcl::PointCloud<pcl::VFHSignature308> > vfh_vec;

      pcl::fromROSMsg (*cloud_msg, *cloud);

      double sac_threshhold;
      ros::param::get("~sac_distence_threshold",sac_threshhold);
      double cluster_tolerance;
      ros::param::get("~cluster_tolerance",cluster_tolerance);
      ex.extractCluster(cloud, cluster_vec, sac_threshhold, cluster_tolerance, 100, 10000);
#ifdef DEBUG
      pcl::ExtractIndices<pcl::PointXYZ> extract;
      extract.setInputCloud (cloud);
      extract.setIndices (ex.inliers_);
      // Remove the planar inliers, extract the rest, which are the clusters
      extract.setNegative (false);
      extract.filter (*cloud_seg);
      sprintf(filename,"plane.pcd");
      pcl::io::savePCDFileASCII (filename, *cloud_seg);

      for(int i=0;i<ex.cluster_number;i++)
      {
          sprintf(filename,"clustter%d.pcd",i);
          pcl::io::savePCDFileASCII (filename, cluster_vec[i]);
          std::cout<<filename<<std::endl;
      }
#endif
      for(int i=0;i<ex.cluster_number;i++)
      {
          *cluster = cluster_vec[i];
          feature.computeOVFH(cluster);
          vfh_vec.push_back(*feature.descriptors_ovfh);
      }
      vnn.estimate6DOF(vfh_vec,cluster_vec);

#ifdef DEBUG
      pcl::io::savePCDFileASCII ("obj.pcd", *vnn.cloud_obj);
      std::cout<<vnn.neighbor_model_name<<std::endl;
#endif
      std_msgs::String neighbor;
      neighbor.data=vnn.neighbor_model_name;
      pub_neighbor.publish(neighbor);

      bool get_name;
      ros::param::get("/cloud_pose/get_neighborname",get_name);
      while(get_name==false)
      {
        ros::Duration(0.5).sleep();
        ros::param::get("/cloud_pose/get_neighborname",get_name);
      }
      sensor_msgs::PointCloud2 obj;
      pcl::toROSMsg (*vnn.cloud_obj, obj);
      pub_obj.publish (obj);

      ros::param::set("~seg_flag",false);
    }



}

int
main (int argc, char** argv)
{
  // Initialize ROS
  ros::init (argc, argv, "segmentation");
  ros::NodeHandle nh;

  pub_obj = nh.advertise<sensor_msgs::PointCloud2> ("cloud_obj", 1);
  pub_neighbor = nh.advertise<std_msgs::String>("neighbor_name", 1);

  // Create a ROS subscriber for the input point cloud
  ros::Subscriber sub = nh.subscribe<sensor_msgs::PointCloud2> ("cloud_remove_nan", 1, cloud_cb);

  while (ros::ok())
  {
      ros::spinOnce();
  }
  return 0;
}
