#include <ros/ros.h>
#include <string.h>
#include "boost/units/io.hpp"	//must---
#include <actionlib/client/simple_action_client.h>
#include <ik/ikAction.h>
#include <geometry_msgs/Pose.h>
#include<geometry_msgs/PoseStamped.h>
#include "brics_actuator/JointPositions.h"
#include "geometry_msgs/Twist.h"
#include <std_msgs/Int32.h>
#include "arm_interface_youbot.h"

using namespace std;
std_msgs::Int32 controllerState;   
ros::Publisher platformPublisher;
ros::Subscriber Marker_Pos_Subscriber;
ros::Publisher statePub;

cArmInterfaceYoubot* pArmInterfaceYoubot;

typedef actionlib::SimpleActionClient<ik::ikAction> Client;

std::vector<double> marker_jointvalues(5);	//存储“目标点”轴值
std::vector<double> adjust_jointvalues(5);	//存储“预抓点”轴值
bool marker_ik = false;  //计算“目标点”是否逆解，定义标志位
bool adjust_ik = false;  //计算“预抓点”是否逆解，定义标志位
void grasp_action();



void mark_pose_callback(const geometry_msgs::PoseStamped &msg)
{
    ROS_INFO("IK solver is comein");
    geometry_msgs::Pose marker_pose;	//定义“目标点”位姿
    geometry_msgs::Pose adjust_pose;	//定义“预抓点”位姿
    marker_pose = msg.pose;  //暂存接收到的mark坐标值

    marker_ik = false;    //计算“目标点”是否逆解，标志位false
    adjust_ik = false;    //计算“目标点”是否逆解，标志位false

//   marker_pose.position.x=0.4039;//0.277672;//0.3139994056;//0.315714;//0.33445;//0.32774;//0.306907;//0.3075;//0.27;
//   marker_pose.position.y=0;//-0.03;//-0.036004;//-0.037275;//-0.0942323;//-0.0106974;//0.0403334;//-0.036842;//0.047693662;
//   marker_pose.position.z=0.00;//0.0852054 ;//0.0928915;//0.0881834;//0.0395986;//0.044775;//0.0144775;//0.01962990031;
    marker_pose.orientation.x=0;//-0.00209345;//0;//-0.00258904;//0.00557639;//-0.0227564;//-0.0276205;//0.0064-0.0121;
    marker_pose.orientation.y=1;//0.997838;//1;//0.0907523;//0.087495;//0.00162876;//0.94906;//0.0951041;//0.998112;//687;
    marker_pose.orientation.z=0;//-0.00778829;//0;//0.00206682;//-0.000272122 ;//0.00322423;//0.0431536;//0.0020.05
    marker_pose.orientation.w=0;//0.0652307;//0;//0.704197;//0.704394;//0.706919;//0.31903;//0.7039 ;//.9984;

    adjust_pose=marker_pose;
    adjust_pose.position.z  = adjust_pose.position.z+0.01;    
	marker_pose.position.z  = marker_pose.position.z-0.04;


    Client ik_client("youbot_ik",true);
    ik_client.waitForServer();
    ik::ikGoal goal_pose;
//---第一次计算---------------------------------------------------
    goal_pose.pose=marker_pose;
    ik_client.sendGoal(goal_pose);
    ik_client.waitForResult(ros::Duration(10.0));

    if(ik_client.getState()==actionlib::SimpleClientGoalState::SUCCEEDED)   //getState()
    {
        //如果逆解成功，则“抓取系列动作”使能
        ROS_INFO("IK1 solver is ok");
   		ik::ikResult p=*(ik_client.getResult());

        //ROS_INFO("have %d joints",p.jointresult.size());

        std::cout<<"have "<<p.jointresult.size()<<" joints"<<std::endl;
   		for(int i=0;i<p.jointresult.size();i++)
   		{
        	std::cout<<p.jointresult[i]<<std::endl;
       		marker_jointvalues[i]=p.jointresult[i];
            //这里需要把joints值，通过变量传出来
            //ROS_INFO("id:%f",msg.markers[0].id);
   		}
        marker_ik = true;
		//	pArmInterfaceYoubot->PublishJointValues(jointvalues);
		//	pArmInterfaceYoubot->PublishGripperValues(0.025);     
		//	moveArm_cont(jointvalues);
		//	std::cout<<"get the ik valve"<<std::endl;
		//使能标志位，手臂抓取动作函数
    }
    else
    {
        marker_ik = false;
		ROS_INFO("IK1 solver is false");
    }
//--第二次计算----------------------------------------------
    goal_pose.pose=adjust_pose;
    ik_client.sendGoal(goal_pose);
    ik_client.waitForResult(ros::Duration(10.0));

    if(ik_client.getState()==actionlib::SimpleClientGoalState::SUCCEEDED)   //getState()
    {
        //如果逆解成功，则“抓取系列动作”使能
        ROS_INFO("IK2 solver is ok");
   		ik::ikResult p=*(ik_client.getResult());

        //ROS_INFO("have %d joints",p.jointresult.size());

    	std::cout<<"have "<<p.jointresult.size()<<" joints"<<std::endl;
 	  	for(int i=0;i<p.jointresult.size();i++)
 	  	{
 	       	std::cout<<p.jointresult[i]<<std::endl;
 	       	adjust_jointvalues[i]=p.jointresult[i];
 	       	//这里需要把joints值，通过变量传出来
           	//ROS_INFO("id:%f",msg.markers[0].id);
   		}
		adjust_ik = true;
    }
    else
    {
        adjust_ik = false;
        ROS_INFO("IK2 solver is false");
    }
	grasp_action();
//---------------------------------------------------------
    ros::Duration(5).sleep();
}



void grasp_action()
{
    if(marker_ik && adjust_ik)
    {
		ROS_INFO("ARM Ik is ok");
//		pArmInterfaceYoubot -> Arm_Grasp(marker_jointvalues,adjust_jointvalues);
        pArmInterfaceYoubot->PublishJointValues(adjust_jointvalues);
    	ros::Duration(5).sleep();
		pArmInterfaceYoubot->OpenGrippers();
		ros::Duration(3).sleep();
        pArmInterfaceYoubot->PublishJointValues(marker_jointvalues);
    	ros::Duration(3).sleep();
		pArmInterfaceYoubot->CloseGrippers();
		ros::Duration(3).sleep();

		pArmInterfaceYoubot->Arm_Place();
		ros::Duration(5).sleep();
		pArmInterfaceYoubot->OpenGrippers();
		ros::Duration(3).sleep();
    }
	else
	{
		controllerState.data = 1;
        statePub.publish(controllerState);
	}
}



int main(int argc,char **argv)
{
	ros::init(argc, argv, "youbot_tag_pick_client");
	ros::NodeHandle n;
    ros::NodeHandle ph("~");
	pArmInterfaceYoubot= new cArmInterfaceYoubot(ph);

        Marker_Pos_Subscriber = n.subscribe("marker_pose",1,&mark_pose_callback);
//	platformPublisher = n.advertise<geometry_msgs::Twist>("cmd_vel", 1);
   	statePub=n.advertise<std_msgs::Int32>("/control_current_state", 1);
	sleep(1);	//must!!!!!  must!!!!!


	pArmInterfaceYoubot -> Arm_Upright();
	ros::Duration(4).sleep();
	pArmInterfaceYoubot -> Arm_Recognition_Ready();
	ros::Duration(4).sleep();
	controllerState.data = 1;
    statePub.publish(controllerState);

	sleep(1);
	ros::spin();
}






