#include <moveit/move_group_interface/move_group.h>
#include <moveit/planning_scene_interface/planning_scene_interface.h>

#include <moveit_msgs/DisplayRobotState.h>
#include <moveit_msgs/DisplayTrajectory.h>

#include <moveit_msgs/AttachedCollisionObject.h>
#include <moveit_msgs/CollisionObject.h>
#include <moveit/robot_model_loader/robot_model_loader.h>

#include <moveit/robot_model/robot_model.h>
#include <moveit/robot_state/robot_state.h>

#include<moveit/kinematics_base/kinematics_base.h>
#include <iostream>
#include <stdlib.h>
#include<string>
#include<std_msgs/String.h>

#include <ros/ros.h>
#include <eigen_conversions/eigen_msg.h>
#include<geometry_msgs/Pose.h>
#include <actionlib/server/simple_action_server.h>
#include <ik/ikAction.h>

typedef actionlib::SimpleActionServer<ik::ikAction> Server;







int main(int argc, char **argv)
{
 ros::init(argc,argv,"youbot_ik");
 ros::NodeHandle n;
 ROS_INFO("start");
 ros::Time init();
      ros::Time t_1,t_2,t_3,t_4,t_5;
      ros::Duration t,t_delta,t_delta2,t_delta3;

      t_3=ros::Time::now();
      robot_model_loader::RobotModelLoader robot_model_loader("robot_description");

     robot_model::RobotModelPtr  kinematic_model=robot_model_loader.getModel();

      ROS_INFO("Model frame:%s",kinematic_model->getModelFrame().c_str());

     robot_state::RobotStatePtr kinematic_state(new robot_state::RobotState(kinematic_model));

      const robot_state::JointModelGroup* joint_model_group=kinematic_model->getJointModelGroup("arm1");

      const std::vector<std::string>&jointName=joint_model_group->getJointModelNames();

      kinematic_state->setToDefaultValues();
      std::vector<double> joint_vaules;
      kinematic_state->copyJointGroupPositions(joint_model_group,joint_vaules);
      for(std::size_t i=0;i<joint_vaules.size();i++)
      {
         std::cout << jointName[i] << " " << joint_vaules[i] << std::endl;
      }

      joint_vaules[0]=1.57;
      kinematic_state->setJointGroupPositions(joint_model_group,joint_vaules);
      std::cout << "Current state is" << (kinematic_state->satisfiesBounds()?"valid":"not valid") << std::endl;
      kinematic_state->enforceBounds();
      std::cout << "Current state is" << (kinematic_state->satisfiesBounds()?"valid":"not valid") << std::endl;

      for(std::size_t i=0;i<joint_vaules.size();i++)
      {
         std::cout << jointName[i] << " " << joint_vaules[i] << std::endl;
      }


      kinematic_state->setToDefaultValues();
      kinematic_state->setToRandomPositions(joint_model_group);
      //const Eigen::Affine3d&end_effector_state = kinematic_state->getGlobalLinkTransform("gripper_palm_link");
      //std::cout << "Translation:" << end_effector_state.translation() << std::endl;
      //std::cout << "Rotation:" << end_effector_state.rotation() << std::endl;
      t_4=ros::Time::now();
      t_delta=t_4-t_3;
      geometry_msgs::Pose q;

      geometry_msgs::Pose tf;
      tf.position.x=0.341859;//0.24056;//0.315714;//0.33445;//0.32774;//0.306907;//0.3075;//0.291829;//0.236127;//0.485805;//0.3092;
      tf.position.y=-0.037275;//-0.0942323;//-0.0106974;//0.0403334;//-0.036842;//0.0476939;//-0.0555437;//-0.0283366;//-0.0268962;//-0.00309964;//0.01536;
      tf.position.z=0.0881834;//0.0395986;//0.044775;//0.0144775;//0.0196299;//0.0057158;//0.0603816;//0.00760306;//0.0416987;//0.0497992;//0.158527;//0.0031;
      tf.orientation.x=-0.00209345;//0;//-0.00258904;//0.00557639;//-0.0227564;//-0.0276205;//0.00646734;//0.0605274;//0.997249;//-0.0248331;//0.0121;
      tf.orientation.y=0.997838;//1;//0.0907523;//0.087495;//0.00162876;//0.94906;//0.0951041;//0.998112;//0.0467639;//0.999687;//-0.0523;
      tf.orientation.z=-0.00778829;//0;//0.00206682;//-0.000272122 ;//0.00322423;//0.0431536;//0.00236107;//-0.000467502;//0.0566458;//-0.000124637;//-0.00092;
      tf.orientation.w=0.0652307;//0;//0.704197;//0.704394;//0.706919;//0.31903;//0.7039 ;//0.010419;//0.00989743;//0.00308141;//0.9984;

      q=tf;

       Eigen::Affine3d end_effector_state ;

      tf::poseMsgToEigen(q,end_effector_state);

      std::cout << "Translation:" << end_effector_state.translation() << std::endl;
      std::cout << "Rotation:" << end_effector_state.rotation() << std::endl;



      t_1=ros::Time::now();
      t_delta2=t_1-t_4;
      bool found_ik=kinematic_state->setFromIK(joint_model_group,end_effector_state,10,0.1);
      //bool found_ik=kinematics::KinematicsBase::getPositionIK(q,);
      if(found_ik)
      {
         kinematic_state->copyJointGroupPositions(joint_model_group,joint_vaules);
         for(std::size_t i=0;i<jointName.size();i++)
         {
            std::cout << jointName[i] << " " << joint_vaules[i] << std::endl;
         }


         ik::ikResult result;

         for(int i=0;i<joint_vaules.size();i++)
         {
             result.jointresult.push_back(joint_vaules[i]);
             std::cout<<result.jointresult[i]<<std::endl;
         }
         t_2=ros::Time::now();
         t=t_2-t_1;


      }
      else
      {
         std::cout << "can't find IK solution." << std::endl;

      }

      std::cout<<"载入模型耗时"<<t_delta<<" "<<t_delta2<<"反解时间 "<<t<<std::endl;



 ros::spin();

}
